from pygame.rect import Rect
from random import random, randint, choice

from model import *
from operator import add

class GameParameters (object):
	"""Structure to keep initial game conditions.

	Should be present before start of the game.
	"""
	start_cash = 50000
	city_fps = 1

	use_city_prices = False

	lynch_fame = -20
	target_fame = 1000
	destroy_fame = 1
	park_fame = 2

	map_w = 128
	map_h = 128
	terrain = MildTerrain()

class Dozer (object):
	def __init__(self, initial, fps, order, map, target, start, queue, callback):
		self.counter = initial
		self.fps = fps
		self.map = map
		self.order = order
		self.target = target
		self.next = start
		self.end = None
		self.queue = queue
		self.callback = callback

		self.position = None
		self.idle = False
	
	def think(self):
		if self.counter == 0:
			self.move()
			self.counter = self.fps
		else:
			self.counter -= 1

	def move(self):
		if self.idle:
			return
	
		self.position = self.queue[self.next]
		was = self.map.get_terrain(self.position[0], self.position[1])
		self.map.set_terrain(self.position[0], self.position[1], self.target.tid)

		if self.callback:
			self.callback(was, self.target.tid)

		self.order.done += 1
		self.next += 1

		if self.next >= self.end:
			self.position = None
			self.idle = True

class Order (object):
	"""Construction order."""

	# Frames per square - dozer running speed.
	fps = 2
	# Squares per dozer - number of dozers to launch.
	spd = 500
	

	def __init__(self, map, rect, target, callback):
		self.rect = rect
		self.target = target
		self.map = map

		self.complete = False
		self.dozers = None
		self.done = 0

		self.on_complete = None

		queue_a = [[(rect.left + x, rect.top + y) for x in range(rect.w)] for y in range(0, rect.h, 2)]
		queue_b = [[(rect.left + x, rect.top + y) for x in range(rect.w - 1, -1, -1)] for y in range(1, rect.h, 2)]
		
		queue = range(rect.h)
		queue[::2] = queue_a
		queue[1::2] = queue_b
		queue = reduce(add, queue)

		# Filter out unbuildable
		queue = [c for c in queue if self.map.get_owner(c[0], c[1]) == 1]

		self.dozers = []
		self.area = len(queue)

		ndozer = self.area / self.spd + 1
		dozer = None
		for d in range(ndozer):
			start = d * self.area / ndozer
			if dozer:
				dozer.end = start

			dozer = Dozer(d, self.fps, self, map, target, start, queue, callback)
			self.dozers.append(dozer)

		dozer.end = self.area

	def think(self):
		for dozer in self.dozers:
			dozer.think()

		if self.done == self.area:
			self.complete = True
			if self.on_complete:
				self.on_complete()
	
	def cancel(self):
		self.complete = True
		self.dozers = []

class NoFunds (Exception):
	pass

class NoGood (Exception):
	pass

class Player (object):
	def __init__(self, game, cash):
		self.game = game
		self.cash = cash
		self.fame = 0

		get_type = self.game.map.terraininfo.get_type
		self.forest = get_type('forest').tid
		self.park = get_type('park').tid
		self.orders = []
		self.ventures = dict.fromkeys(self.game.map.types.keys(), 0)

	def update_ventures(self, was, now):
		# This will get non-buildable types into negative, but we don't care
		# since they don't have any income.
		self.ventures[was] -= 1
		self.ventures[now] += 1

		types = self.game.map.terraininfo.types
		twas = types[was]
		tnow = types[now]

		if tnow.natural != twas.name:
			self.fame -= self.game.params.destroy_fame
		
		if was != self.park and now == self.park:
			self.fame += self.game.params.park_fame

	def think(self):
		for o in self.orders:
			o.think()
		self.orders = [o for o in self.orders if not o.complete]

	def payday(self):
		income = reduce(add, [self.game.map.types[t].income * self.ventures[t] for t in self.ventures])
		self.got_cash(income)
		self.cash += income

	def create_order(self, rect, target):
		rinfo, land_cost, nogood = self.game.map.get_rect_info(rect)
		cost = self.game.map.get_project(target.tid, rinfo)
		
		if nogood:
			raise NoGood

		if cost > self.cash:
			raise NoFunds

		self.cash -= cost

		self.game.map.set_owner(rect, 1)
		o = Order(self.game.map, rect, target, self.update_ventures)
		self.orders.append(o)

		return o

	def buy_land(self, rect):
		rinof, land_cost, nogood = self.game.map.get_rect_info(rect)
		if nogood:
			raise NoGood

		if land_cost > self.cash:
			raise NoFunds

		self.cash -= land_cost
		self.game.map.set_owner(rect, 1)


class City (object):
	def __init__(self, game, fps):
		self.game = game
		self.fps = fps
		self.map = self.game.map
		self.type = self.map.terraininfo.get_type('city').tid
		self.fronts = []
		self.build_first_block()
		self.counter = 0
	
	def build_first_block(self):
		x = self.map.w - 1
		y = self.map.h - 1
	
		self.map.set_terrain_owner(x, y, self.type, 2)
		self.fronts.append((x, y))

	def can_to(self, pos):
		x, y = pos
		return (0 <= x < self.map.w and 0 <= y < self.map.h and
				self.map.get_terrain(x, y, self.type) != self.type and
				self.map.get_owner(x, y) == 0)

	def expand(self, front_id):
		pos = self.fronts[front_id]
		neigh = [d.calc(*pos) for d in DIRECTIONS]
		avail = [p for p in neigh if self.can_to(p)]
		if not avail:
			del self.fronts[front_id]
			return
		next = choice(avail)

		self.map.set_terrain_owner(next[0], next[1], self.type, 2)
		if self.game.params.use_city_prices:
			self.map.increment_cities(*next)
		if len(avail) == 1:
			del self.fronts[front_id]

		self.fronts.append(next)

	def started_over(self):
		pass
	
	def start_over(self):
		nx = randint(0, self.map.w - 1)
		ny = randint(0, self.map.h - 1)
		if self.can_to((nx, ny)):
			self.started_over()
			self.fronts.append((nx, ny))
			self.expand(0)
		
	def think(self):
		if self.counter == 0:
			self.counter = self.fps
			if self.fronts:
				self.expand(randint(0, len(self.fronts)-1))
			else:
				self.start_over()

		else:
			self.counter -= 1
		
class Date (object):
	days = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
	day_names = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']

	def __init__(self, year, month, day):
		self.year = year
		self.month = month
		self.day = day
		self.hours = 0
		self.minutes = 0
		self.wday = 0
	
	def inc(self, minutes = 10):
		self.minutes += minutes

		while self.minutes >= 60:
			self.hours += 1
			self.minutes -= 60

		while self.hours >= 24:
			self.day += 1
			self.wday += 1
			self.hours -= 24

		while self.day > self.days[(self.month - 1) % len(self.days)]:
			self.month += 1
			self.day = 1
		
		while self.month > 12:
			self.year += 1
			self.month = 1

	def __str__(self):
		return '%s %02d.%02d.%04d' % (self.day_names[self.wday % 7], self.day, self.month, self.year)

class Game (object):
	"""Toplevel gameplay class.

	Instantiated when the game starts.
	"""
	def __init__(self, params=GameParameters):
		self.params = params

		self.map = Map(params.map_w, params.map_h, params.terrain)
		self.date = Date(2009, 4, 1)

		self.player = Player(self, params.start_cash)
		self.computer = City(self, params.city_fps)
		self.next_payday = 7

		self.over = False
	
	def game_over(self, why):
		self.over = why
	
	def think(self):
		self.date.inc()
		if self.date.wday == self.next_payday:
			self.player.payday()
			self.next_payday += 7

		if self.player.cash < 0:
			self.game_over('bankrupt')

		if self.player.fame < self.params.lynch_fame:
			self.game_over('lynched')
		
		if self.player.fame > self.params.target_fame:
			self.game_over('won')

		self.player.think()
		self.computer.think()

